
yd, xd = self.output_dims
galimage = np.zeros((yd, xd))
segmentation = segmap.SegMap()
        segmentation.xdim = xd
        segmentation.ydim = yd
        segmentation.initialize_map()
for source in blah:
    galdims = stamp.shape
            psf_image = self.create_psf_stamp(entry['pixelx'], entry['pixely'], (galdims[1], galdims[0]), 1)

            xpts, ypts, (i1, i2), (j1, j2), (k1, k2), (l1, l2) = self.create_psf_stamp_coords(entry['pixelx'],
                                                                                              entry['pixely'],
                                                                                              (galdims[1],
                                                                                               galdims[0]))
            stamp = s1.fftconvolve(stamp, psf_image, mode='same')

























        psfimage = np.zeros(self.output_dims)

        for source in point_sources:
        # PTSRC CODE FOR COMPARISON -- REMOVE WHEN DONE
            xpos_aperture = entry['pixelx'] + self.coord_adjust['xoffset']
            ypos_aperture = entry['pixely'] + self.coord_adjust['yoffset']
            xpos_fullframe = entry['pixelx'] + self.subarray_bounds[0]
            ypos_fullframe = entry['pixely'] + self.subarray_bounds[1]
            #nx = math.floor(xpos)
            #ny = math.floor(ypos)

            # Desired counts per second in the point source
            counts = entry['countrate_e/s']  # / self.frametime

            # Load PSF from the appropriate file given the source location
            # Remember that this source location should be in the coordinate
            # system of the aperture being simulated, and not include any padding
            # for WFSS seed images.

            # UPDATED FOR NEW PSF LIBRARY---------------------------------

            # Separate the coordinates of the PSF into rounded integer and fractional pixels.
            # Work in full frame coordinates. Adjust if simulating a subarray.
            #subpix_x, integx = np.modf(entry['pixelx'] + self.subarray_bounds[0])
            #subpix_y, integy = np.modf(entry['pixely'] + self.subarray_bounds[1])

            # Interpolate the oversampled PSF to the nearest whole pixel
            # This creates a 2d interpolated PSF
            print("CLEAN UP HERE")
            #psf_integer_interp = self.psf_library.position_interpolation(integx, integy, method='idw')

            # FOR USE WITH OLD PSF LIBRARY---------------------------------
            # psf_obj = PSF(entry['pixelx'], entry['pixely'], self.psfname,
            #              interval=self.params['simSignals']['psfpixfrac'], oversampling=1)

            # Calculate the coordinate limits of the aperture/PSF stamp overlap
            # psf_ydim, psf_xdim = psf_obj.model.shape
            #psf_xdim, psf_ydim = (self.psf_library_x_dim, self.psf_library_y_dim)
            (i1, i2, j1, j2, k1, k2, l1, l2) = self.cropped_coords(xpos_aperture, ypos_aperture,
                                                                   self.psf_library_x_dim,
                                                                   self.psf_library_y_dim,
                                                                   self.output_dims[1], self.output_dims[0])
            (i1_ff, i2_ff, j1_ff, j2_ff, k1_ff, k2_ff, l1_ff, l2_ff) = self.cropped_coords(xpos_fullframe, ypos_fullframe,
                                                                                           self.psf_library_x_dim,
                                                                                           self.psf_library_y_dim,
                                                                                           self.ffsize, self.ffsize)

            #try:
            if 1>0:
                ypts, xpts = np.mgrid[j1:j2, i1:i2]
                ypts_ff, xpts_ff = np.mgrid[j1_ff:j2_ff, i1_ff:i2_ff]

                # FOR USE WITH OLD PSF LIBRARY
                # scaled_psf = psf_obj.model.evaluate(x=xpts, y=ypts, flux=counts, x_0=nx, y_0=ny)

                # FOR USE WITH NEW PSF LIBRARY
                scaled_psf = self.psf_library.evaluate(x=xpts_ff, y=ypts_ff, flux=counts, x_0=xpos_fullframe,
                                                       y_0=xpos_fullframe)
                #scaled_psf = psf_integer_interp.evaluate(x=xpts, y=ypts, flux=counts, x_0=xpos, y_0=ypos)
                psfimage[ypts, xpts] += scaled_psf
        # PTSRC CODE FOR COMPARISON -- REMOVE WHEN DONE


#! /usr/bin/env python

"""
Extract a roughly SCA-sized area from a distortion-free mosaic.
This extracted area can then be blotted to match NIRCam distorion
and then used as an input to the ramp simulator.
"""

import argparse

from astropy import wcs
from astropy.io import fits
import numpy as np
from jwst import datamodels

class Extraction():
    def __init__(self, mosaicfile=None, data_extension_number=0, wcs_extension_number=0,
                 center_ra=None, center_dec=None, outfile=None, dimensions=(0, 0),
                 jwst_pixel_scale=None):
        """
        Parameters
        ----------
        mosaicfile : str
            Name of the fits file containing the image to be cropped

        data_extension_number : int
            Extension number within the mosaic fits file where the data are stored

        wcs_extension_number : int
            Extension number within the mosaic fits file where the WCS information
            are stored

        center_ra : float
            RA value in decimal degrees at the cropped image's reference location

        center_dec : float
            Dec value in decimal degrees at the cropped image's reference location

        outfile : str
            Name of fits file to save the cropped image into. If None, the image is
            not saved, but is still available in self.cropped

        dimensions : tup
            Tuple of (x, y) dimensions, in pixels, of the sub-image to be cropped
            from the mosaic. These are pixels in the pixel scale of the JwST
            aperture that the cropped image will be used for

        jwst_pixel_scale : float
            Pixel scale in arcsec/pixel of the JWST aperture which the cropped
            image will be used for later. This is needed to transform the
            ``dimensions`` from JWST pixels to mosaic pixels.
        """
        self.mosaicfile = mosaicfile
        self.data_extension_number = data_extension_number
        self.wcs_extension_number = wcs_extension_number
        self.center_ra = center_ra
        self.center_dec = center_dec
        self.outfile = outfile
        self.dimensions = dimensions
        self.jwst_pixel_scale = jwst_pixel_scale

        # Buffer around the extraction dimensions to include.
        # This is necessary in case the cropped image will be
        # rotated during later resampling, in order to avoid
        # having the rotated image corners fall outside the
        # extracted image. This is a multiplicative value.
        self.buffer = np.sqrt(2.)


    def extract(self):
        """MAIN FUNCTION: extract a subimage from the input mosaic
        """
        # Get WCS info from mosaic file
        mosaic = fits.open(self.mosaicfile)
        mosaic_wcs = wcs.WCS(mosaic[self.wcs_extension_number].header)
        mosaic_shape = mosaic[self.data_extension_number].data.shape
        try:
            self.instrument = mosaic[0].header['INSTRUME']
        except KeyError:
            self.instrument = 'NA'

        # Assume the input mosaic has been drizzled, so remove
        # any SIP coefficients that may be present
        print('\n\n******************************************************************')
        print('Assuming input files have been drizzled. Removing SIP coefficients')
        print('******************************************************************\n\n')
        mosaic_wcs.sip = None

        radec = np.array([[self.center_ra, self.center_dec]])
        mosaic_xy_at_center_radec = mosaic_wcs.wcs_world2pix(radec, 1)
        mosaic_center_x = mosaic_xy_at_center_radec[0][0] - 1
        mosaic_center_y = mosaic_xy_at_center_radec[0][1] - 1

        # These are only for populating the header of the cropped file
        mosaic_ra_ref, mosaic_dec_ref = mosaic_wcs.wcs.crval
        mosaic_x_ref, mosaic_y_ref = mosaic_wcs.wcs.crpix
        self.mosaic_x_type, self.mosaic_y_type = mosaic_wcs.wcs.ctype
        self.cd11, self.cd12 = mosaic_wcs.wcs.cd[0]
        self.cd21, self.cd22 = mosaic_wcs.wcs.cd[1]
        self.mosaic_scale_x = np.abs(self.cd11) * 3600.  # arcsec/pix
        self.mosaic_scale_y = np.abs(self.cd22) * 3600.  # arcsec/pix
        self.mosaic_roll = np.arccos(self.cd11 / (self.mosaic_scale_x / 3600.))

        # Define size of region to extract
        # self.dimensions are in units of jwst pixwls
        #outscale_x = self.nrc_scale[self.channel.lower()]
        xlen, ylen = self.dimensions

        # Expand by the buffer
        xlen *= self.buffer
        ylen *= self.buffer

        # Determine the dimensions of the mosaic aperture to be cropped in
        # units of mosaic pixels
        nx = np.absolute(np.int(xlen * self.jwst_pixel_scale / self.mosaic_scale_x))
        ny = np.absolute(np.int(ylen * self.jwst_pixel_scale / self.mosaic_scale_y))

        # Set the CRPIX values that will define the WCS in the output
        self.crpix1 = nx / 2 + 0.5
        self.crpix2 = ny / 2 + 0.5

        # Extract the SCA-sized area from the moasic
        half_height = ny // 2 + 1
        half_width = nx // 2 + 1

        miny = np.int(mosaic_center_y - half_height)
        maxy = np.int(mosaic_center_y + half_height + 1)
        minx = np.int(mosaic_center_x - half_width)
        maxx = np.int(mosaic_center_x + half_width + 1)

        # If the cropped area falls off the edge of the mosaic, adjust
        # the extraction coordinates and the crpix values accordingly
        if miny < 0:
            self.crpix2 += miny
            miny = 0
        if minx < 0:
            self.crpix1 += minx
            minx = 0
        if maxy > mosaic_shape[0]:
            diff = maxy - mosaic_shape[0]
            maxy = mosaic_shape[0]
        if maxx > mosaic_shape[1]:
            maxx = mosaic_shape[1]

        print("Coords of center of cropped area", mosaic_center_x, mosaic_center_y)
        print("X-min, X-max coords: ", minx, maxx)
        print("Y-min, Y-max coords: ", miny, maxy)

        crop = mosaic[self.data_extension_number].data[miny: maxy, minx: maxx]

        # Place into a data model to prepare for blotting
        self.cropped = self.populate_datamodel(crop)

        # Save only if outfile is not None
        if self.outfile is not None:
            self.savefits(crop, self.outfile)
            print("Extracted image saved to {}".format(self.outfile))

    def populate_datamodel(self, array):
        """Place the image and accopanying WCS information in an
        ImageModel instance. This makes passing the information to
        the blotting function easier.

        Parameters
        ----------
        array : numpy.ndarray
            2D array containing the cropped image
        """

        # now we need to update the WCS of the mosaic piece
        cropped = datamodels.ImageModel()
        cropped.meta.wcsinfo.cdelt1 = self.mosaic_scale_x / 3600.
        cropped.meta.wcsinfo.cdelt2 = self.mosaic_scale_y / 3600.
        cropped.meta.wcsinfo.crpix1 = self.crpix1
        cropped.meta.wcsinfo.crpix2 = self.crpix2
        cropped.meta.wcsinfo.crval1 = self.center_ra
        cropped.meta.wcsinfo.crval2 = self.center_dec
        cropped.meta.wcsinfo.ctype1 = self.mosaic_x_type
        cropped.meta.wcsinfo.ctype2 = self.mosaic_y_type
        cropped.meta.wcsinfo.cunit1 = 'deg'
        cropped.meta.wcsinfo.cunit2 = 'deg'
        cropped.meta.wcsinfo.ra_ref = self.center_ra
        cropped.meta.wcsinfo.dec_ref = self.center_dec
        cropped.meta.wcsinfo.roll_ref = self.mosaic_roll * 180./np.pi
        cropped.meta.wcsinfo.wcsaxes = 2
        cropped.meta.wcsinfo.pc1_1 = self.cd11 / (self.mosaic_scale_x / 3600.)
        cropped.meta.wcsinfo.pc1_2 = self.cd12 / (self.mosaic_scale_x / 3600.)
        cropped.meta.wcsinfo.pc2_1 = self.cd21 / (self.mosaic_scale_y / 3600.)
        cropped.meta.wcsinfo.pc2_2 = self.cd22 / (self.mosaic_scale_y / 3600.)

        cropped.data = array
        return cropped

    def savefits(self, array, ofile):
        """Save the given array in a fits file

        Parameters
        ----------
        array : numpy.ndarray
            2D array containing the data to be saved

        ofile : str
            Output filename to use
        """
        h0 = fits.PrimaryHDU()
        h1 = fits.ImageHDU()
        h1.data = array

        h1.header['INSTRUME'] = self.instrument
        h1.header['CDELT1'] = self.mosaic_scale_x / 3600.
        h1.header['CDELT2'] = self.mosaic_scale_y / 3600.
        h1.header['CRPIX1'] = self.crpix1
        h1.header['CRPIX2'] = self.crpix2
        h1.header['CRVAL1'] = self.center_ra
        h1.header['CRVAL2'] = self.center_dec
        h1.header['CTYPE1'] = self.mosaic_x_type
        h1.header['CTYPE2'] = self.mosaic_y_type
        h1.header['CUNIT1'] = 'deg'
        h1.header['CUNIT2'] = 'deg'
        h1.header['RA_REF'] = self.center_ra
        h1.header['DEC_REF'] = self.center_dec
        h1.header['ROLL_REF'] = self.mosaic_roll * 180./np.pi
        h1.header['WCSAXES'] = 2
        h1.header['EXTNAME'] = 'SCI    '
        h1.header['EXTVER'] = 1
        h1.header['PC1_1'] = self.cd11 / (self.mosaic_scale_x / 3600.)
        h1.header['PC1_2'] = self.cd12 / (self.mosaic_scale_x / 3600.)
        h1.header['PC2_1'] = self.cd21 / (self.mosaic_scale_y / 3600.)
        h1.header['PC2_2'] = self.cd22 / (self.mosaic_scale_y / 3600.)

        hlist = fits.HDUList([h0, h1])
        hlist.writeto(ofile, overwrite=True)

    def add_options(self, parser=None, usage=None):
        if parser is None:
            parser = argparse.ArgumentParser(usage=usage, description='Extract SCA-sized area from moasic')
        parser.add_argument("--mosaicfile", help="Filename of fits file containing mosaic.")
        parser.add_argument("--data_extension_number", help="Extension number in mosaic file where data are located. Default=0", default=0)
        parser.add_argument("--center_ra", help="RA at the center of the extracted area", type=np.float)
        parser.add_argument("--center_dec", help="Dec at the center of the extracted area", type=np.float)
        parser.add_argument("--outfile", help="Name of output fits file containing extracted image", default=None)
        parser.add_argument("--dimensions", help="Tuple of (x pixels, y pixels) of the image to crop. Excludes buffer for WFSS.")
        parser.add_argument("--jwst_pixel_scale", help=("Pixel scale (arcsec/pix) of the detector/aperture "
                                                        "for JWST aperture data will be applied to."))
        return parser


if __name__ == '__main__':
    usagestring = 'USAGE: crop_mosaic.py filename.fits 23.43 -21.2'

    ex = Extraction()
    parser = ex.add_options(usage = usagestring)
    args = parser.parse_args(namespace=ex)
    ex.extract()




#! /usr/bin/env python

"""This module contains functions for creating Mirage-formatted source
catalogs

maybe i should change self.ra to self._ra and same with self.dec, to encourage
people to use the get_ra() and get_dec() functions

Right now create_table returns a table, which is fine, but would also be nice
to be able to do:

tab = PointSourceCatalog(stuff)
tab_ap_format = tab.create_table()
tab.save() - and have tab.save save the tab_ap_format table


Maybe:
tab = PointSourceCatalog(stuff)
tab.create_table()  # creates tab.astropy_table or something like that
tab.save()

"""

import copy
import os

from astropy.io import ascii
from astropy.table import Table, Column
import numpy as np

TSO_GRISM_INDEX = 99999

class PointSourceCatalog():
    def __init__(self, ra=[], dec=[], x=[], y=[]):
        """Initialize the point source catalog. Users can enter lists of RA and Dec values
        or x and y values for source potisions

        Parameters
        ----------
        something
        """
        # Make sure we are working with numpy arrays
        ra = np.array(ra)
        dec = np.array(dec)
        x = np.array(x)
        y = np.array(y)

        if len(ra) != len(dec):
            raise ValueError(("WARNING: inconsistent length between RA and Dec inputs."))
        if len(x) != len(y):
            raise ValueError(("WARNING: inconsistent length between x and y inputs."))
        if len(ra) > 0 and len(x) > 0:
            raise ValueError(("WARNING: Provide either RA and Dec values, or x and y values."))

        if len(ra) > 0:
            self._ra = ra
            self._dec = dec
        else:
            self._ra = x
            self._dec = y
        self.magnitudes = {}

        # Determine the units for the location fields. All that Mirage needs to know is whether
        # the units are pixels or not. Degrees vs hour angle is not important at this point.
        if len(ra) > 0:
            self._location_units = 'position_RA_Dec'
        if len(x) > 0:
            self._location_units = 'position_pixels'

    def add_catalog(self, catalog_to_add, magnitude_fill_value=99.):
        """Add a catalog to the current catalog instance"""
        # If the the source positions in the two catalogs have different units, then the catalogs
        # can't be combined.
        if self._location_units != catalog_to_add.location_units:
            raise ValueError("WARNING: Sources in the catalogs do not have matching units (RA/Dec or x/y)")

        # Check the magnitude system from each and make sure they match
        current_mag_labels = list(self.magnitudes.keys())
        new_mag_labels = list(catalog_to_add.magnitudes.keys())
        mag_sys = self.magnitudes[current_mag_labels[0]][0]
        if mag_sys != catalog_to_add.magnitudes[new_mag_labels[0]][0]:
            print("WARNING: Magnitude systems of the two catalogs do not match. Cannot combine.")

        # Get the length of the two catalogs
        current_length = len(self._ra)
        new_length = len(catalog_to_add._ra)

        # Combine location columns
        self._ra = np.append(self._ra, catalog_to_add._ra)
        self._dec = np.append(self._dec, catalog_to_add._dec)

        # Now we need to compare magnitude columns. Columns common to both catalogs can be
        # combined. Columns not common will have to have fill values added so that everything
        # is the same length
        orig_current_mag_labels = copy.deepcopy(current_mag_labels)
        current_mag_labels.extend(new_mag_labels)
        mag_label_set = set(current_mag_labels)
        for label in mag_label_set:
            if ((label in orig_current_mag_labels) and (label not in new_mag_labels)):
                fill = [magnitude_fill_value] * new_length
                self.magnitudes[label][1] = np.append(self.magnitudes[label][1], fill)
            if ((label not in orig_current_mag_labels) and (label in new_mag_labels)):
                fill = [magnitude_fill_value] * current_length
                self.magnitudes[label] = [mag_sys, np.append(fill, catalog_to_add.magnitudes[label][1])]
            if ((label in orig_current_mag_labels) and (label in new_mag_labels)):
                self.magnitudes[label][1] = np.append(self.magnitudes[label][1],
                                                      catalog_to_add.magnitudes[label][1])

        # Update the catalog table if it already exists, so that it is consistent
        self.create_table()

    def add_magnitude_column(self, magnitude_list, magnitude_system='abmag', instrument='', filter_name=''):
        """Add a list of magnitudes to the catalog

        Parameters
        ----------
        some_stuff
        """
        # Force magnitude list to be a numpy array
        magnitude_list = np.array(magnitude_list)

        # Make sure instrument and filter are allowed values
        instrument = instrument.lower()
        filter_name = filter_name.lower()
        self.filter_check(instrument, filter_name)

        # Create the string for the column header
        if instrument == '' or filter_name == '':
            header = 'magnitude'
        else:
            if instrument.lower() in ['fgs1', 'fgs2']:
                header = '{}_magnitude'.format(instrument.lower())
            else:
                header = '{}_{}_magnitude'.format(instrument, filter_name)

        # Get a list of magnitude_system for the existing catalog. No mixing of
        # magnitude systems is currently allowed.
        keys = self.magnitudes.keys()
        if len(keys) > 0:
            current_mag_sys = self.magnitudes[list(keys)[0]][0]
        else:
            current_mag_sys = magnitude_system

        # Make sure this is not a duplicate column
        if header in self.magnitudes.keys():
            raise ValueError(("WARNING: {} entry already exists in catalog. No duplicate entries allowed."
                              .format(header)))
        # Make sure the magnitude system is consistent
        elif magnitude_system != current_mag_sys:
                raise ValueError(("WARNING: only a single magnitude system is allowed per catalog. "
                                  "Current catalog is using {}, while the new entry is {}."
                                  .format(current_mag_sys, magnitude_system)))
        else:
            self.magnitudes[header] = [magnitude_system, magnitude_list]

        # Update the catalog table
        self.create_table()

    def filter_check(self, inst_name, filt_name):
        """Make sure the requested instrument/filter pair is valid
        """
        if inst_name == 'nircam':
            filter_file = 'nircam_filter_pupil_pairings.list'
        elif inst_name == 'niriss':
            filter_file = 'niriss_dual_wheel_list.txt'

        if inst_name in ['nircam', 'niriss']:
            filter_file_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '../config/',
                                                            filter_file))
            filter_table = ascii.read(filter_file_path)
            if filt_name.upper() not in filter_table['filter']:
                raise ValueError("WARNING: {} is not a valid filter for {}.".format(filt_name,
                                                                                    inst_name.upper()))
        if inst_name == 'fgs':
            if filt_name.lower() not in ['guider1', 'guider2']:
                raise ValueError("WARNING: {} is not a valid filter for FGS.".format(filt_name))

    def instrument_check(self, inst_name):
        """Only certain instrument values are allowed
        NOT USED - so that we can return J, H, K from searches
        """
        if inst_name not in ['nircam', 'niriss', 'fgs']:
            raise ValueError("WARNING: {} is not a valid instrument.".format(inst_name))

    @property
    def dec(self):
        """Return Dec values from catalog"""
        if self._location_units == 'position_RA_Dec':
            return self._dec
        else:
            return []

    def get_magnitudes(self, key):
        """Return the magnitude list (but not mag_sys) associated with the given key

        Parameters
        ----------
        justone
        """
        try:
            return self.magnitudes[key][1]
        except KeyError:
            print("WARNING: No {} magnitude column present.".format(key))

    @property
    def location_units(self):
        """Return RA values from catalog"""
        return self._location_units

    @property
    def ra(self):
        """Return RA values from catalog"""
        if self._location_units == 'position_RA_Dec':
            return self._ra
        else:
            return []

    @property
    def x(self):
        """Return x values from catalog"""
        if self._location_units == 'position_pixels':
            return self._ra
        else:
            return []

    @property
    def y(self):
        """Return y values from catalog"""
        if self._location_units == 'position_pixels':
            return self._dec
        else:
            return []

    def create_table(self):
        """Create an astropy table containing the catalog
        """
        tab = create_basic_table(self._ra, self._dec, self.magnitudes, self._location_units)

        # Make sure there are at least 4 comment lines at the top
        self.table = pad_table_comments(tab)

    def save(self, output_name):
        """Write out the catalog to an ascii file

        Parameters
        ----------
        something
        """
        self.table.write(output_name, format='ascii', overwrite=True)


class GalaxyCatalog(PointSourceCatalog):
    def __init__(self, ra=[], dec=[], x=[], y=[], ellipticity=[], radius=[], sersic_index=[],
                 position_angle=[], radius_units='arcsec'):
        """Create instance of a galaxy catalog
        """
        # Add location information
        PointSourceCatalog.__init__(self, ra=ra, dec=dec, x=x, y=y)

        # Add galaxy-specific information
        self._morphology = {'radius': radius, 'ellipticity': ellipticity, 'sersic_index': sersic_index,
                            'pos_angle': position_angle}

        self._ellipticity = ellipticity
        self._pos_angle = position_angle
        self._radius = radius
        self._sersic_index = sersic_index

        if radius_units not in ['arcsec', 'pixels']:
            raise ValueError("WARNING: Galaxy radii must be in units of 'arcsec' or 'pixels'.")
        self._radius_units = radius_units

    def create_table(self):
        """Create an astropy table containing the catalog
        """
        tab = create_basic_table(self._ra, self._dec, self.magnitudes, self.location_units)

        # Add morphology columns
        for key in self._morphology:
            values = self._morphology[key]
            col = Column(values, name=key)
            tab.add_column(col, index=3)

        # Add magnitude system and position units as metadata
        tab.meta['comments'].append("radius_{}".format(self.radius_units))

        # Make sure there are at least 4 comment lines at the top
        self.table = pad_table_comments(tab)

    @property
    def ellipticity(self):
        """Return Dec values from catalog"""
        return self._ellipticity

    @property
    def position_angle(self):
        """Return Dec values from catalog"""
        return self._pos_angle

    @property
    def radius(self):
        """Return Dec values from catalog"""
        return self._radius

    @property
    def sersic_index(self):
        """Return Dec values from catalog"""
        return self._sersic_index

    @property
    def radius_units(self):
        """Return Dec values from catalog"""
        return self._radius_units

    @property
    def morphology(self):
        """Return Dec values from catalog"""
        return self._morphology


class ExtendedCatalog(PointSourceCatalog):
    def __init__(self, filenames=[], ra=[], dec=[], x=[], y=[], position_angle=[]):
        # Add location information
        PointSourceCatalog.__init__(self, ra=ra, dec=dec, x=x, y=y)

        # Add extended source-specific information
        self._filenames = filenames
        self._pos_angle = position_angle

    def create_table(self):
        """Create an astropy table containing the catalog
        """
        tab = create_basic_table(self._ra, self._dec, self.magnitudes, self._location_units)

        # Add the filename column
        file_col = Column(self._filenames, name='filename')
        tab.add_column(file_col, index=3)

        # Add the position angle column
        pa_col = Column(self._pos_angle, name='pos_angle')
        tab.add_column(pa_col)

        # Make sure there are at least 4 comment lines at the top
        self.table = pad_table_comments(tab)

    @property
    def filename(self):
        """Return Dec values from catalog"""
        return self._filenames

    @property
    def position_angle(self):
        """Return Dec values from catalog"""
        return self._pos_angle


class MovingPointSourceCatalog(PointSourceCatalog):
    def __init__(self, ra=[], dec=[], x=[], y=[], ra_velocity=[], dec_velocity=[], x_velocity=[],
                 y_velocity=[]):
        # Add location information
        PointSourceCatalog.__init__(self, ra=ra, dec=dec, x=x, y=y)

        if len(ra_velocity) > 0 and len(x_velocity) > 0:
            raise ValueError(("WARNING: Provide either RA and Dec velocities, or x and y "
                              "velocities, but not both."))
        # Object velocities
        if len(ra_velocity) > 0:
            self._ra_velocity = ra_velocity
            self._dec_velocity = dec_velocity
            self._velocity_units = 'velocity_RA_Dec'
        else:
            self._ra_velocity = x_velocity
            self._dec_velocity = y_velocity
            self._velocity_units = 'velocity_pixels'

    def create_table(self):
        tab = create_basic_velocity_table(self._ra, self._dec, self.magnitudes, self._location_units,
                                          self._ra_velocity, self._dec_velocity, self._velocity_units)

        # Make sure there are at least 4 comment lines at the top
        self.table = pad_table_comments(tab)

    @property
    def dec_velocity(self):
        """Return Dec values from catalog"""
        return self._dec_velocity

    @property
    def ra_velocity(self):
        """Return Dec values from catalog"""
        return self._ra_velocity

    @property
    def velocity_units(self):
        """Return Dec values from catalog"""
        return self._velocity_units


class MovingSersicCatalog(GalaxyCatalog, MovingPointSourceCatalog):
    def __init__(self, ra=[], dec=[], x=[], y=[], ra_velocity=[], dec_velocity=[], x_velocity=[],
                 y_velocity=[], ellipticity=[], radius=[], sersic_index=[], position_angle=[],
                 radius_units='arcsec'):
        # Add location information
        MovingPointSourceCatalog.__init__(self, ra=ra, dec=dec, x=x, y=y, ra_velocity=ra_velocity,
                                          dec_velocity=dec_velocity, x_velocity=x_velocity,
                                          y_velocity=y_velocity)
        GalaxyCatalog.__init__(self, ra=ra, dec=dec, x=x, y=y, ellipticity=ellipticity, radius=radius,
                               sersic_index=sersic_index, position_angle=position_angle,
                               radius_units='arcsec')

    def create_table(self):
        tab = create_basic_velocity_table(self._ra, self._dec, self.magnitudes, self._location_units,
                                          self._ra_velocity, self._dec_velocity, self._velocity_units)
        # Add morphology columns
        for key in self._morphology:
            values = self._morphology[key]
            col = Column(values, name=key)
            tab.add_column(col, index=3)

        # Add magnitude system and position units as metadata
        tab.meta['comments'].append("radius_{}".format(self._radius_units))

        # Make sure there are at least 4 comment lines at the top
        self.table = pad_table_comments(tab)


class MovingExtendedCatalog(ExtendedCatalog, MovingPointSourceCatalog):
    def __init__(self, ra=[], dec=[], x=[], y=[], ra_velocity=[], dec_velocity=[], x_velocity=[],
                 y_velocity=[], filenames=[], position_angle=[]):
        # Add location information
        MovingPointSourceCatalog.__init__(self, ra=ra, dec=dec, x=x, y=y, ra_velocity=ra_velocity,
                                          dec_velocity=dec_velocity, x_velocity=x_velocity,
                                          y_velocity=y_velocity)
        ExtendedCatalog.__init__(self, ra=ra, dec=dec, x=x, y=y, filenames=filenames,
                                 position_angle=position_angle)

    def create_table(self):
        tab = create_basic_velocity_table(self._ra, self._dec, self.magnitudes, self._location_units,
                                          self._ra_velocity, self._dec_velocity, self._velocity_units)
        # Add filename column
        file_col = Column(self._filenames, name='filename')
        tab.add_column(file_col, index=1)

        # Add position_angle column
        pa_col = Column(self._pos_angle, name='pos_angle')
        tab.add_column(pa_col, index=6)

        # Make sure there are at least 4 comment lines at the top
        self.table = pad_table_comments(tab)


class NonSiderealCatalog(MovingPointSourceCatalog):
    def __init__(self, ra=[], dec=[], x=[], y=[], ra_velocity=[], dec_velocity=[], x_velocity=[],
                 y_velocity=[], object_type=[]):
        # Add location information
        MovingPointSourceCatalog.__init__(self, ra=ra, dec=dec, x=x, y=y, ra_velocity=ra_velocity,
                                          dec_velocity=dec_velocity, x_velocity=x_velocity,
                                          y_velocity=y_velocity)

        # List of the type of sources in the non-sidereal catalog
        valid_objects = ['pointSource', 'galaxies', 'extended']
        for obj in object_type:
            if obj not in valid_objects:
                raise ValueError(("WARNING: object_type list members must be one of: {}"
                                  .format(obj, valid_objects)))
        self.object_type = object_type

    def create_table(self):
        """Create an astropy table containing the catalog
        """
        tab = create_basic_velocity_table(self._ra, self._dec, self.magnitudes, self._location_units,
                                          self._ra_velocity, self._dec_velocity, self._velocity_units)
        obj_col = Column(self.object_type, name='object')
        tab.add_column(obj_col, index=1)

        # Make sure there are at least 4 comment lines at the top
        self.table = pad_table_comments(tab)


class ImagingTSOCatalog(PointSourceCatalog):
    def __init__(self, ra=[], dec=[], x=[], y=[], lightcurve_file=[]):
        """Create instance of an imaging TSO catalog
        """
        # Add location information
        PointSourceCatalog.__init__(self, ra=ra, dec=dec, x=x, y=y)

        # Add TSO-specific information
        self._lightcurve_file = lightcurve_file

    def create_table(self):
        """Create an astropy table containing the catalog
        """
        tab = create_basic_table(self._ra, self._dec, self.magnitudes, self.location_units,
                                 minimum_index=TSO_GRISM_INDEX)

        # Add the lightcurve filename column
        lc_col = Column(self._lightcurve_file, name='lightcurve_file')
        tab.add_column(lc_col, index=3)

        # Make sure there are at least 4 comment lines at the top
        self.table = pad_table_comments(tab)

    @property
    def lightcurve_file(self):
        """Return lightcurve filenames"""
        return self._lightcurve_file


class GrismTSOCatalog(PointSourceCatalog):
    def __init__(self, ra=[], dec=[], x=[], y=[], semimajor_axis=[], orbital_inclination=[],
                 eccentricity=[], orbital_period=[], longitude_of_periastron=[], limb_dark_model=[],
                 limb_dark_coeffs=[], time_units=[], start_time=[], end_time=[], inferior_conj=[],
                 transmission_spectrum=[]):
        """Create instance of a grism TSO catalog
        """
        # Add location information
        PointSourceCatalog.__init__(self, ra=ra, dec=dec, x=x, y=y)

        # Add TSO-specific information
        self._semimajor_axis = np.array(semimajor_axis)
        self._orbital_inclination = np.array(orbital_inclination)
        self._eccentricity = np.array(eccentricity)
        self._orbital_period = np.array(orbital_period)
        self._longitude_of_periastron = np.array(longitude_of_periastron)
        self._limb_dark_model = np.array(limb_dark_model)
        self._time_units = np.array(time_units)
        self._start_time = np.array(start_time)
        self._end_time = np.array(end_time)
        self._inferior_conj = np.array(inferior_conj)
        self._transmission_spectrum = np.array(transmission_spectrum)

        # Limb darkening coefficients should be a string of comma-separated
        # numbers
        str_limb_dark_coeffs = []
        for entry in limb_dark_coeffs:
            str_limb_dark_coeffs.append(str(entry).replace('[', '').replace(']', ''))
        self._limb_dark_coeffs = np.array(str_limb_dark_coeffs)

    def create_table(self):
        """Create an astropy table containing the catalog
        """
        tab = create_basic_table(self._ra, self._dec, self.magnitudes, self.location_units,
                                 minimum_index=TSO_GRISM_INDEX)

        # Add the lightcurve filename column
        semimajor_col = Column(self._semimajor_axis, name='Semimajor_axis_in_stellar_radii')
        tab.add_column(semimajor_col, index=3)

        orb_inc_col = Column(self._orbital_inclination, name='Orbital_inclination_deg')
        tab.add_column(orb_inc_col, index=4)

        ecc_col = Column(self._eccentricity, name='Eccentricity')
        tab.add_column(ecc_col, index=5)

        pastron_col = Column(self._longitude_of_periastron, name='Longitude_of_periastron')
        tab.add_column(pastron_col, index=6)

        limb_model_col = Column(self._limb_dark_model, name='Limb_darkening_model')
        tab.add_column(limb_model_col, index=7)

        limb_coeff_col = Column(self._limb_dark_coeffs, name='Limb_darkening_coeffs')
        tab.add_column(limb_coeff_col, index=8)

        time_unit_col = Column(self._time_units, name='Time_units')
        tab.add_column(time_unit_col, index=9)

        start_col = Column(self._start_time, name='Start_time')
        tab.add_column(start_col, index=10)

        end_col = Column(self._end_time, name='End_time')
        tab.add_column(end_col, index=11)

        inf_conj_col = Column(self._inferior_conj, name='Time_of_inferior_conjunction')
        tab.add_column(inf_conj_col, index=12)

        orb_per_col = Column(self._orbital_period, name='Orbital_period')
        tab.add_column(orb_per_col, index=13)

        spec_col = Column(self._transmission_spectrum, name='Transmission_spectrum')
        tab.add_column(spec_col, index=14)

        # Make sure there are at least 4 comment lines at the top
        self.table = pad_table_comments(tab)

    @property
    def semimajor_axis(self):
        return self._semimajor_axis

    @property
    def orbital_inclination(self):
        return self._orbital_inclination

    @property
    def eccentricity(self):
        return self._eccentricity

    @property
    def longitude_of_periastron(self):
        return self._longitude_of_periastron

    @property
    def limb_dark_model(self):
        return self._limb_dark_model

    @property
    def limb_dark_coeffs(self):
        return self._limb_dark_coeffs

    @property
    def time_units(self):
        return self._time_units

    @property
    def start_time(self):
        return self._start_time

    @property
    def end_time(self):
        return self._end_time

    @property
    def inferior_conj(self):
        return self._inferior_conj

    @property
    def orbital_period(self):
        return self._orbital_period

    @property
    def transmission_spectrum(self):
        return self._transmission_spectrum


def add_velocity_columns(input_table, ra_velocities, dec_velocities, velocity_units):
    """
    DO IT
    """
    ra_vel_col = Column(ra_velocities, name='x_or_RA_velocity')
    dec_vel_col = Column(dec_velocities, name='y_or_Dec_velocity')
    input_table.add_columns([ra_vel_col, dec_vel_col], indexes=[3, 3])
    input_table.meta['comments'].append(velocity_units)
    return input_table


def cat_from_file(filename, catalog_type='point_source'):
    """Read in a Mirage-formatted ascii catalog file and place into an
    instance of a catalog object

    Parameters
    ----------
    filename : str
        Name of ascii catalog to be read in

    catalog_type : str
        Type of source catalog. Allowed values are:
        point_source, galaxy, extended, moving_point_source, moving_sersic,
        moving_extended, non_sidereal

    Returns
    -------
    catalog_object : mirage.catalogs.catalog_generator object
    """
    catalog_type = catalog_type.lower()
    allowed_types = ['point_source', 'galaxy', 'extended', 'moving_point_source',
                     'moving_sersic', 'moving_extended', 'non_sidereal']
    if catalog_type not in allowed_types:
        raise ValueError(("Input catalog type {} is not one of the allowed types: {}"
                         .format(catalog_type, allowed_types)))

    cat_table = ascii.read(filename)

    if 'position_pixels' in cat_table.meta['comments'][0:4]:
        xpos = 'x'
        ypos = 'y'
    else:
        xpos = 'ra'
        ypos = 'dec'

    magnitude_columns = [col for col in cat_table.colnames if 'magnitude' in col]

    # Point Source catalog
    if catalog_type == 'point_source':
        catalog_object = PointSourceCatalog(**{xpos: cat_table['x_or_RA']}, **{ypos: cat_table['y_or_Dec']})
        for magcol in magnitude_columns:
            instrument, filtername = get_inst_filter_from_colname(magcol)
            catalog_object.add_magnitude_column(cat_table[magcol], instrument=instrument,
                                                filter_name=filtername)

    # Galaxy catalog
    elif catalog_type == 'galaxy':
        radius_units = 'arcsec'
        if 'radius_pixels' in cat_table.meta['comments'][0:4]:
            radius_units = 'pixels'
        catalog_object = GalaxyCatalog(**{xpos: cat_table['x_or_RA']}, **{ypos: cat_table['y_or_Dec']},
                                       ellipticity=cat_table['ellipticity'], radius=cat_table['radius'],
                                       sersic_index=cat_table['sersic_index'], position_angle=cat_table['pos_angle'],
                                       radius_units=radius_units)
        for magcol in magnitude_columns:
            instrument, filtername = get_inst_filter_from_colname(magcol)
            catalog_object.add_magnitude_column(cat_table[magcol], instrument=instrument,
                                                filter_name=filtername)

    # Extended catalog
    elif catalog_type == 'extended':
        catalog_object = ExtendedCatalog(**{xpos: cat_table['x_or_RA']}, **{ypos: cat_table['y_or_Dec']},
                                         filenames=cat_table['filename'], position_angle=cat_table['pos_angle'])
        for magcol in magnitude_columns:
            instrument, filtername = get_inst_filter_from_colname(magcol)
            catalog_object.add_magnitude_column(cat_table[magcol], instrument=instrument,
                                                filter_name=filtername)

    # Moving point source catalog
    elif catalog_type == 'moving_point_source':
        velocity_units = 'velocity_RA_Dec'
        xvel = 'ra_velocity'
        yvel = 'dec_velocity'
        if 'velocity_pixels' in cat_table.meta['comments'][0:4]:
            velocity_units = 'velocity_pixels'
            xvel = 'x_velocity'
            yvel = 'y_velocity'
        catalog_object = MovingPointSourceCatalog(**{xpos: cat_table['x_or_RA']},
                                                  **{ypos: cat_table['y_or_Dec']},
                                                  **{xvel: cat_table['x_or_RA_velocity']},
                                                  **{yvel: cat_table['y_or_Dec_velocity']})
        for magcol in magnitude_columns:
            instrument, filtername = get_inst_filter_from_colname(magcol)
            catalog_object.add_magnitude_column(cat_table[magcol], instrument=instrument,
                                                filter_name=filtername)

    # Moving sersic catalog
    elif catalog_type == 'moving_sersic':
        velocity_units = 'velocity_RA_Dec'
        xvel = 'ra_velocity'
        yvel = 'dec_velocity'
        if 'velocity_pixels' in cat_table.meta['comments'][0:4]:
            velocity_units = 'velocity_pixels'
            xvel = 'x_velocity'
            yvel = 'y_velocity'
        radius_units = 'arcsec'
        if 'radius_pixels' in cat_table.meta['comments'][0:4]:
            radius_units = 'pixels'
        catalog_object = MovingSersicCatalog(**{xpos: cat_table['x_or_RA']},
                                             **{ypos: cat_table['y_or_Dec']},
                                             **{xvel: cat_table['x_or_RA_velocity']},
                                             **{yvel: cat_table['y_or_Dec_velocity']},
                                             ellipticity=cat_table['ellipticity'],
                                             radius=cat_table['radius'],
                                             sersic_index=cat_table['sersic_index'],
                                             position_angle=cat_table['pos_angle'],
                                             radius_units=radius_units)
        for magcol in magnitude_columns:
            instrument, filtername = get_inst_filter_from_colname(magcol)
            catalog_object.add_magnitude_column(cat_table[magcol], instrument=instrument,
                                                filter_name=filtername)

    # Moving extended catalog
    elif catalog_type == 'moving_extended':
        velocity_units = 'velocity_RA_Dec'
        xvel = 'ra_velocity'
        yvel = 'dec_velocity'
        if 'velocity_pixels' in cat_table.meta['comments'][0:4]:
            velocity_units = 'velocity_pixels'
            xvel = 'x_velocity'
            yvel = 'y_velocity'
        catalog_object = MovingExtendedCatalog(**{xpos: cat_table['x_or_RA']},
                                               **{ypos: cat_table['y_or_Dec']},
                                               **{xvel: cat_table['x_or_RA_velocity']},
                                               **{yvel: cat_table['y_or_Dec_velocity']},
                                               filenames=cat_table['filename'],
                                               position_angle=cat_table['pos_angle'])
        for magcol in magnitude_columns:
            instrument, filtername = get_inst_filter_from_colname(magcol)
            catalog_object.add_magnitude_column(cat_table[magcol], instrument=instrument,
                                                filter_name=filtername)

    # Non-sidereal catalog
    elif catalog_type == 'non_sidereal':
        velocity_units = 'velocity_RA_Dec'
        xvel = 'ra_velocity'
        yvel = 'dec_velocity'
        if 'velocity_pixels' in cat_table.meta['comments'][0:4]:
            velocity_units = 'velocity_pixels'
            xvel = 'x_velocity'
            yvel = 'y_velocity'
        catalog_object = NonSiderealCatalog(**{xpos: cat_table['x_or_RA']},
                                            **{ypos: cat_table['y_or_Dec']},
                                            **{xvel: cat_table['x_or_RA_velocity']},
                                            **{yvel: cat_table['y_or_Dec_velocity']},
                                            object_type=cat_table['object'])
        for magcol in magnitude_columns:
            instrument, filtername = get_inst_filter_from_colname(magcol)
            catalog_object.add_magnitude_column(cat_table[magcol], instrument=instrument,
                                                filter_name=filtername)
    return catalog_object


def create_basic_table(ra_values, dec_values, magnitudes, location_units, minimum_index=1):
    """Create astropy table containing the basic catalog info
    NOTE THAT THIS IS OUTSIDE CLASSES
    """
    basic_table = Table()

    # Add index, filename, RA, Dec or x, y columns
    index_col = Column(np.arange(minimum_index, minimum_index + len(ra_values)), name='index')
    ra_col = Column(ra_values, name='x_or_RA')
    dec_col = Column(dec_values, name='y_or_Dec')
    basic_table.add_columns([index_col, ra_col, dec_col])

    # Add magnitude columns
    for key in magnitudes:
        mag_values = magnitudes[key][1]
        mag_sys = magnitudes[key][0]
        mag_column = Column(mag_values, name=key)
        basic_table.add_column(mag_column)

    # Add magnitude system and position units as metadata
    basic_table.meta['comments'] = [location_units, mag_sys]
    return basic_table


def create_basic_velocity_table(ra_values, dec_values, magnitudes, location_units,
                                ra_velocities, dec_velocities, velocity_units):
    tab = create_basic_table(ra_values, dec_values, magnitudes, location_units)
    tab = add_velocity_columns(tab, ra_velocities, dec_velocities, velocity_units)
    return tab


def get_inst_filter_from_colname(column_name):
    """Get the instrument and filter name from a magnitude column string

    Parameters
    ----------
    column_name : str
        Magnitude column header from catalog (e.g. 'nircam_f444w_magnitude')

    Returns
    -------
    instrument : str
        Instrument name

    filten_name : str
        Filter name
    """
    parts = column_name.split('_')
    instrument = parts[0].lower()
    if instrument == 'guider':
        filter_name = ''
    else:
        filter_name = parts[1].lower()
    return instrument, filter_name


def pad_table_comments(input_table):
    """do it"""
    if len(input_table.meta['comments']) < 4:
        pad = [''] * (4 - len(input_table.meta['comments']))
        input_table.meta['comments'].extend(pad)
    return input_table



'''Define unit tests for calculation of single frame exposure times

Authors
-------
    - Bryan Hilbert

Use
---
    Ensure you have pytest installed. Then, simply run pytest in any
    parent directory of mirage/tests/:
    >>> pytest
'''

import numpy as np

from mirage.utils.utils import calc_frame_time


def test_fgs_cal_frametime():
    """These tests apply to the FGS CAL apertures only. ACQ1, ACQ2, ID,
    TRK, and FG are not yet supported
    """
    fgs_full = calc_frame_time('fgs', 'FGS_', 2048, 2048, 4)
    assert np.isclose(fgs_full, 10.73677, rtol=0., atol=1e-5)

    fgs_128 = calc_frame_time('fgs', 'FGS_', 128, 128, 1)
    assert np.isclose(fgs_128, 0.1820, rtol=0., atol=1e-5)

    fgs_32 = calc_frame_time('fgs', 'FGS_', 32, 32, 1)
    assert np.isclose(fgs_32, 0.01254, rtol=0., atol=1e-5)

    fgs_8 = calc_frame_time('fgs', 'FGS_', 8, 8, 1)
    assert np.isclose(fgs_8, 0.00126, rtol=0., atol=1e-5)


def test_nircam_frametime():
    """Test NIRCam exposure times
    """
    nrc_full = calc_frame_time('nircam', 'NRCA1_FULL', 2048, 2048, 4)
    assert np.isclose(nrc_full, 10.73677, rtol=0., atol=1e-5)

    nrc_640 = calc_frame_time('nircam', 'NRCA1_SUB640', 640, 640, 1)
    assert np.isclose(nrc_640, 4.18584, rtol=0., atol=1e-5)

    nrc_320 = calc_frame_time('nircam', 'NRCA1_SUB320', 320, 320, 1)
    assert np.isclose(nrc_320, 1.06904, rtol=0., atol=1e-5)

    nrc_160 = calc_frame_time('nircam', 'NRCA1_SUB160', 160, 160, 1)
    assert np.isclose(nrc_160, 0.27864, rtol=0., atol=1e-5)

    nrc_64 = calc_frame_time('nircam', 'NRCB4_SUB64P', 64, 64, 1)
    assert np.isclose(nrc_64, 0.05016, rtol=0., atol=1e-5)

    nrc_32 = calc_frame_time('nircam', 'NRC_SUB32TATS', 32, 32, 1)
    assert np.isclose(nrc_32, 0.01496, rtol=0., atol=1e-5)

    nrc_subgrism256_1 = calc_frame_time('nircam', 'NRC_SUBGRISM256', 2048, 256, 1)
    print(nrc_subgrism256_1, 5.31480)
    #assert np.isclose(nrc_subgrism256_1, 5.29420, rtol=0., atol=1e-5)

    nrc_subgrism256_4 = calc_frame_time('nircam', 'NRC_SUBGRISM256', 2048, 256, 4)
    print(nrc_subgrism256_4, 1.34669)
    #assert np.isclose(nrc_subgrism256_4, 1.34669, rtol=0., atol=1e-5)

    nrc_subgrism128_1 = calc_frame_time('nircam', 'NRC_SUBGRISM128', 2048, 128, 1)
    print(nrc_subgrism128_1, 2.67800)
    #assert np.isclose(nrc_subgrism128_1, 2.6574, rtol=0., atol=1e-5)

    nrc_subgrism128_4 = calc_frame_time('nircam', 'NRC_SUBGRISM128', 2048, 128, 4)
    assert np.isclose(nrc_subgrism128_4, 0.67597, rtol=0., atol=1e-5)

    nrc_subgrism64_1 = calc_frame_time('nircam', 'NRC_SUBGRISM64', 2048, 64, 1)
    assert np.isclose(nrc_subgrism64_1, 1.35960, rtol=0., atol=1e-5)

    nrc_subgrism64_4 = calc_frame_time('nircam', 'NRC_SUBGRISM64', 2048, 64, 4)
    assert np.isclose(nrc_subgrism64_4, 0.34061, rtol=0., atol=1e-5)


def test_niriss_frametime():
    """Test NIRISS exposure times
    """
    nis_full = calc_frame_time('niriss', 'NIS_CEN', 2048, 2048, 4)
    assert np.isclose(nis_full, 10.73677, rtol=0., atol=1e-5)

    nis_80 = calc_frame_time('niriss', 'NIS_SUB80', 80, 80, 1)
    assert np.isclose(nis_80, 0.07544, rtol=0., atol=1e-5)

    nis_64 = calc_frame_time('niriss', 'NIS_SUBTAAMI', 64, 64, 1)
    assert np.isclose(nis_64, 0.05016, rtol=0., atol=1e-5)

    nis_wfss64r = calc_frame_time('niriss', 'NIS_WFSS64R', 64, 2048, 4)
    assert np.isclose(nis_wfss64r, 0.34061, rtol=0., atol=1e-5)

    nis_wfss64c = calc_frame_time('niriss', 'NIS_WFSS64C', 2048, 64, 1)
    assert np.isclose(nis_wfss64c, 1.55800, rtol=0., atol=1e-5)

    nis_wfss128r = calc_frame_time('niriss', 'NIS_WFSS128R', 128, 2048, 4)
    assert np.isclose(nis_wfss128r, 0.67597, rtol=0., atol=1e-5)

    nis_wfss128c = calc_frame_time('niriss', 'NIS_WFSS128C', 2048, 128, 1)
    assert np.isclose(nis_wfss128c, 2.87000, rtol=0., atol=1e-5)
